/* mandelbrot tile
fractal -zone x0,y0,x1,y1 -size HxW -ofile file ...
minimum memory varsion. Allocate only one line
*/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "cmap.h"

#define IS_CNUM(X) (NULL != strchr("-.0123456789",(X)))

#define ERROR   -1

/* colour map. Mike Futcher, http://yogyog.org */

char   *cmap [] = {
"       s None  c None",
".      c #000000",
"1      c #000081",
"2      c #0000FF",
"3      c #008100",
"4      c #008181",
"5      c #00FF00",
"6      c #00FFFF",
"7      c #810000",
"8      c #810081",
"9      c #818100",
"A      c #818181",
"B      c #87D6AE",
"C      c #FF0000",
"D      c #FF00FF",
"E      c #FFFF00",
"F      c #FFFFFF"};

char  **ycmap = &cmap[0];
int cmap_size = 17;

int j_orbit(), j3_orbit(), j4_orbit(), j5_orbit(), j6_orbit(),
    j_julia(), j_tricorn(), j_burn(), j_wis(), j_bes(),
    nr_cubic(), nr_ecc();


struct JT {int fc; int (*func)(); char *desc; } jt[] = {
{0, j_orbit,  "Mandelbrot z = z^2+c "},
{1, j3_orbit, "Multibrot 3"},
{2, j4_orbit, "Multibrit 4"},
{3, j5_orbit, "Multibrot 5"},
{4, j6_orbit, "Multibrot 6"},
{5, j_julia,  "Julia(a) z = z^2+a z[0]=c"},
{6, j_tricorn, "Tricorn z = (zbar)^2+c"},
{7, j_burn,   "Burning Ship   z = (|Re(z)|+i|Im(z)|)^2+c"},
{8, j_wis,    "Newton Raphson. w = e^z+a*z"},
{9, j_bes,    "Newton Raphson. w = j(0,x) Bessel function"},
{10, nr_cubic, "Newton Raphson. w = x^3+ax+b"},
{11, nr_ecc,   "Newton Raphson. w = y^y-x^3+ax+b"},
{-1, NULL, NULL}};

int     fcmax   = 11;

char   *version =
       "Version: fractal, Sun Dec 20 2009";
char   *info = "(C)Tony Goddard http://d4maths.lowtech.org";


char   *help_text[] = {
"Usage: %s  [options ... number ... number ]\n",
"-help          print options",
"-size HxW      Image size. Rows x columns",
"-zone xmin,ymin,xmax,ymax    Zone. Bounding box",
"-box  xmin,ymin,xmax,ymax    Bounding box. 4 real numbers",
"-bailout integer      Maximum Iterations. Default 34",
"-lut string    Colors to use, normally  .123456789ABCDEF",
"               Colors may be repeated",
"-Lnc           Create a lut with nc colors",
"-eps number    Break NR iteration when |z[n+1]-z[n]|<eps",
"-o string      Output file. If file has .xpm extension create image file",
"               Otherwise write raw data HxW matrix of bytes from lut",
"               When output file is omitted write to standard output",
"-fnn           function code nn",
"number         paramater(s) for designated function code.",
NULL};

double  zw[4]   = {-2.1,-1.5,1.1,1.5};
double  xa[6]   = {0, -1, 0, 0, 0, 0};  /* paramaters */
int     npar    = 0;
int     size[2] = {24,80};
char    xlut[129];              /* single byte color */
int     cnum    = 0;            /* get colormap from cmap.h */
int     kf      = 0;            /* default to mandelbrot */
double  eps     =  1.0e-10;     /* -eps epsilon */
int     bailout = 35;           /* maximum iteration */
char   *ofile   = NULL;
char    lut80[]   =\
".123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ\
abcdefghijklmnopqrstuvwxyz_0\
%&():*+-/=[]{|}~";

char   *lut     = &lut80[0];

int     nlut = 16;

main(argc, argv)
int argc;
char **argv;
{
char   *thisprog = argv[0];
FILE   *ofp = NULL;
char   *s;
char   *rup, *t;
int     i, j, k;
double  x, y, dx, dy;

for(argv++, argc--; 0 < argc--; argv++) {
    s = *argv;
    if (s[0] == '-') switch(s[1]) {
        case 'h':   /* help */
                usage (thisprog);
                return 0;
        case 'b':       /* -bailout */
                if (s[2] != 'o') {
                    bailout = atoi(*++argv);
                    argc--;
                    break;
                    }
        case 'z':   /* -box or -zone */
                k = parse_dbl(4, &zw[0], *++argv);
                if (k != 4) {
                    usage (thisprog);
                    return 0;
                    }
                argc --;
        break;
        case 'C':   /* -Cnn select colormap */
                cnum = atoi(s+2);
                break;
        case 'e':       /* -eps */
                eps = atof(*++argv);
                argc--;
        break;
        case 'f':       /* -fNN function selection */
                kf = atoi(s+2);
                if (kf > fcmax || kf < 0 ) kf = 0;
        break;
        case 'l':       /* lut */
                lut = *++argv;
                nlut = strlen(lut);
                argc--;
        break;
        case 'L':       /* -Lnn create lut of length nn */
                nlut = atoi(s+2);
                if (nlut < 2) nlut = 2;
                else if (nlut > 80) nlut = 80;
                break;
        case 'o':       /* ofile */
                ofile = *++argv;
                argc--;
        break;
        case 's':   /* size */
                k = parse_int(2, &size[0], *++argv);
                if (k != 2) {
                    usage (thisprog);
                    return 0;
                    }
                argc--;
        break;
        default:        /* negative number */
                if (npar < 6) {
                    xa[npar++] = atof(s);
                    }
        break;
        }
    else if (npar < 6) {        /* ordinary number */
         xa[npar++] = atof(s);
         }
    }

if (ofile) {
    ofp = fopen(ofile, "w");
    if (ofp == NULL) return ERROR;
    if (NULL != strstr(ofile, ".xpm")) {
        set_colormap (cnum);
        return do_xpm(ofp);
        }
    }
else    ofp = stdout;

/* create raw matrix of single byte pixel values */

if (NULL == (rup = malloc(1+size[1]))) return ERROR;
dy=(zw[3]-zw[1])/(size[0]-1);
dx=(zw[2]-zw[0])/(size[1]-1);
for (i=0, y=zw[1]; i < size[0]; i++, y +=dy) {
    for (t=rup, j=0, x=zw[0]; j < size[1]; j++, x += dx) {
        k = (jt[kf].func)(y,x);
       *t++ = (k == 0) ? '.': lut[k % nlut];
        }
    for (t = rup, j=0; j < size[1]; j++) fputc(*t++, ofp);
    fputc('\n', ofp);
    }
if (ofile) fclose (ofp);
return 0;
}

int     usage(name)     /* help message */
char   *name;
{
char  **txt;
int     i;
printf ("%s\n", version);
printf ("%s\n", info);
printf ("\n");
printf (help_text[0], name);
for (txt = & help_text[1]; *txt != NULL; txt++) {
    printf("%s\n", *txt);
    }
for (i=0; 0 <= jt[i].fc; i++) {
    printf("%2d %s\n", jt[i].fc, jt[i].desc);
    }
return 0;
}

set_colormap(idx)
int idx;
{
char    *t;
char    **p;
int     i, c;
switch (idx) {
        case 32:        /* cmap 32 */
                ycmap = cmap32;
                cmap_size = sizeof(cmap32)/sizeof(char *);
        break;
        case 64:
                ycmap = cmap64;
                cmap_size = sizeof(cmap64)/sizeof(char *);
        break;
        case 80:
                ycmap = cmap80;
                cmap_size = sizeof(cmap80)/sizeof(char *);
        break;
        default:
                return 0;
        break;
        }
for (i=0, t = &xlut[0], p = ycmap; i < cmap_size; i++, p++) {
    c = (*p)[0];        /* colour table byte */
    if (c != ' ') *t++ = c;     /* skip transparent colour */
    }
*t = 0;         /* terminating byte */
lut = xlut;
nlut = strlen(lut);
return 0;
}

int parse_int(n, dst, src)
int     n;      /* tokens */
int    *dst;    /* place to store int */
char   *src;    /* null terminated string */
{
char  xx[0X81];
int     nt = 0;
char   *s, *t;
int     k, mode;
strncpy (&xx[0], src, 0X80);
xx[0X80] = 0;
for (mode =0, t = &xx[0]; *t; t++) {
    k = IS_CNUM (*t);
    switch (2*mode+k) {
    case 0:     break;
    case 1:     /* find a numeric character */
        s = t;  mode = 1;
        break;
    case 2:     /* find a non numeric character */
        *t = 0;
        if (nt >= n) break;
        dst[nt++] = atoi(s);
        mode = 0;
    case 3:     break;  /* do nothing */
        }
    }
if (mode == 1 && nt < n ) dst[nt++] = atoi(s);
return nt;
}

int parse_dbl(n, dst, src)
int     n;      /* tokens */
double *dst;   /* place to store double */
char   *src;    /* null terminated string */
{
char  xx[0X81];
int     nt = 0;
char   *s, *t;
int     k, mode;
strncpy (&xx[0], src, 0X80);
xx[0X80] = 0;
for (mode =0, t = &xx[0]; *t; t++) {
    k = IS_CNUM (*t);
    switch (2*mode+k) {
    case 0:     break;
    case 1:     /* find a numeric character */
        s = t;  mode = 1;
        break;
    case 2:     /* find a non numeric character */
        *t = 0;
        if (nt >= n) break;
        dst[nt++] = atof(s);
        mode = 0;
    case 3:     break;  /* do nothing */
        }
    }
if (mode == 1 && nt < n ) dst[nt++] = atof(s);
return nt;
}

int     do_xpm(fp)
FILE   *fp;
{
char   *rup, *t;
int     i, j, k;
double  dx, dy, x, y;

fprintf(fp, "/* XPM */\n");
fprintf(fp, "/* -fc %d -ab %g,%g -zone %g,%g,%g,%g -eps %g */\n",
            kf, xa[0], xa[1], zw[0], zw[1], zw[2], zw[3], eps);
fprintf(fp, "static char * fractal_xpm[] = {\n");
fprintf(fp, "\"%d %d %d 1\",\n", size[1], size[0], cmap_size);
fprintf(fp, "/* colors */\n");
for (i=0; i < cmap_size ; i++) {
    fprintf(fp, "\"%s\",\n", ycmap[i]);
    }
if (NULL == (rup = malloc(1+size[1]))) return ERROR;
fprintf(fp, "/* pixels */\n");
dy=(zw[3]-zw[1])/(size[0]-1);
dx=(zw[2]-zw[0])/(size[1]-1);
for (i=0, y=zw[1]; i < size[0]; i++, y +=dy) {
    for (t=rup, j=0, x=zw[0]; j < size[1]; j++, x += dx) {
        k = (jt[kf].func)(y,x);
       *t++ = lut[k % nlut];
        }
    if (i < size[0]-1) {
        fprintf(fp, "\"%s\",\n", rup);
        }
    else {
        fprintf(fp, "\"%s\"};\n", rup);
        }
    }
free(rup);
fclose (fp);
return 0;
}

/*      dynamic system z<-z^2+c with global parameters eps, bailout */

int     j_orbit(cy, cx)
double  cy, cx;
{
double  x = 0.0, y = 0.0;
double  x1, y1, x2, y2;
int     j;
for(j = 0; j < bailout; j++) {
    x2= x*x; y2=y*y;
    if (4.0 < x2+y2) return j;
    x1 = cx + x2-y2;
    y1 = cy + 2*x*y;
    x = x1;
    y = y1;
    }
return 0;
}

/*      dynamic system z<-z^3+c  */                                                 

int     j3_orbit(cy, cx)                                                            
double  cy, cx;                                                                     
{                                                                                   
double  x = 0.0, y = 0.0;
double  x1, y1, x2, y2;
int     j;
for(j = 0; j < bailout; j++) {                                                      
    x2= x*x; y2=y*y;                                                                
    if (4.1 < x2+y2) return j;                                                      
    x1 = cx + x*(x2-3*y2);                                                          
    y1 = cy + y*(3*x2-y2);
    x = x1;                                                                         
    y = y1;
    }                                                                               
return 0;
}

/*      dynamic system z<-z^4+c  */                                                 
                                                                                    
int     j4_orbit(cy, cx)
double  cy, cx;                                                                     
{                                                                                   
double  x = 0.0, y = 0.0;                                                           
double  x1, y1, x2, y2;
int     j;                                                                          
for(j = 0; j < bailout; j++) {
    x2= x*x; y2=y*y;                                                                
    if (4.1 < x2+y2) return j;                                                      
    x1 = cx + x2*x2+y2*y2-6*x2*y2;                                                  
    y1 = cy + 4*x*y*(x2-y2);                                                        
    x = x1;
    y = y1;                                                                         
    }
return 0;                                                                           
}

/*      dynamic system z<-z^5+c  */
                                                                                    
int     j5_orbit(cy, cx)                                                            
double  cy, cx;
{                                                                                   
double  x = 0.0, y = 0.0;                                                           
double  x1, y1, x2, y2, x2y2, x4, y4;                                               
int     j;
for(j = 0; j < bailout; j++) {                                                      
    x2= x*x; y2=y*y; x2y2=x2*y2; x4=x2*x2; y4=y2*y2;
    if (4.1 < x2+y2) return j;                                                      
    x1 = cx + x*(x2*x2+5*y2*y2-10*x2y2);                                            
    y1 = cy + y*(5*x4-10*x2y2+y4);                                                  
    x = x1;                                                                         
    y = y1;
    }                                                                               
return 0;
}                                                                                   

/*      dynamic system z<-z^6+c  */

int     j6_orbit(cy, cx)                                                            
double  cy, cx;                                                                     
{
double  x = 0.0, y = 0.0;                                                           
double  x1, y1, x2, y2, x2y2, x4, y4;                                               
int     j;
for(j = 0; j < bailout; j++) {
    x2= x*x; y2=y*y; x2y2=x2*y2; x4=x2*x2; y4=y2*y2;
    if (4.1 < x2+y2) return j;
    x1 = cx + x2*x2*(x2-16*y2)+y2*y2*(15*x2-y2);
    y1 = cy + x*y*(6*(x4+y4)-20*x2y2);
    x = x1;
    y = y1;
    }
return 0;
}

/*      dynamic system z<-z^2+a with starting point c */

int     j_julia(cy, cx)
double  cy, cx;
{
double  x = cx, y = cy;
double  x1, y1, x2, y2;
int     j;
for(j = 0; j < bailout; j++) {
    x2= x*x; y2=y*y;
    if (4.0 < x2+y2) return j;
    x1 = xa[0] + x2-y2;
    y1 = xa[1] + 2*x*y;
    x = x1;
    y = y1;
    }
return 0;
}


/*  Tricorn fractal z<-(conjugate z)^2+c */

int     j_tricorn(cy, cx)
double  cy, cx;
{
double  x = cx, y = cy;
double  x1, y1, x2, y2;
int     j;
for(j = 0; j < bailout; j++) {
    x2= x*x; y2=y*y;
    if (4.0 < x2+y2) return j;
    x1 = cx + x2-y2;
    y1 = cy - 2*x*y;
    x = x1;
    y = y1;
    }
return 0;
}

/*  Burning ship z<-(|Re(z)|+i|Im(z)|)^2+c */

int     j_burn(cy, cx)
double  cy, cx;
{
double  x = cx, y = cy;
double  x1, y1, x2, y2;
int     j;
for(j = 0; j < bailout; j++) {
    x2= x*x; y2=y*y;
    if (4.0 < x2+y2) return j;
    x1 = cx + x2-y2;
    y1 = cy + 2*fabs(x)*fabs(y);
    x = x1;
    y = y1;
    }
return 0;
}

/*      j_wis newton raphson for w = lambda*z+exp(z)            */
/*      iterate z <- z-f(z)/f'(z) until |f(z)/f'(z)|<eps        */
/*      f(z) = lambda*z+exp(z)-w                                */
/*      f'(z) = lambda+exp(z)                                   */
/*      here we set z = x+iy so exp(z)=exp(x)*cos(y)+i sin(y)   */
/*      here f(0)=1-w and f'(0)=1+lambda                        */
/*      so start the iteration with cy, cx                      */


int     j_wis(cy, cx)
double  cy, cx;
{
double  exp(), cos(), sin();
double  x = cx, y = cy;
double  a, b, c, d, delta, ex, ezr, ezi, dx, dy;
double  lambda;
int     j;
lambda = (npar > 0) ? xa[0] : 0.35;
for(j = 0; j < bailout; j++) {
    if (10000 < x*x+y*y) return 0;      /* infinity */
    ex = exp(x);  ezr = ex*cos(y); ezi = ex*sin(y);
    delta = lambda*lambda+ex*ex+2*lambda*ezr;
    a = lambda*x-cx+ezr;
    b = ezi-cy;
    c = lambda + ezr;
/*  d = ezi          */
    dx = (a*c + b*ezi)/delta;
    dy = (b*c - a*ezi)/delta;
    if ((dx*dx+dy*dy) < eps*eps) return j;
    x = x - dx;
    y = y - dy;
    }
return 0;
}

/* solve w=j(z) by z[n+1]=z[n]-(j(z[n])-w)/j'(z[n])     */
/* compute j(z), j'(z) by series                        */
/* j(z) = sum {(z^2n)/4^n*(n!)^2 : n=0, infinity}       */
/* j(lambda, z) =                                       */
/* sum {z^2n/ 4^n*n! * product(lambda+i:i=1..n): n=0..infinity */

int     j_bes(cy, cx)
double  cy, cx;
{
int     j_bbox();
double  x = cx, y = cy;
double  a,b,c,d, delta, dx, dy;
int     j, k;
for(j = 0; j < bailout; j++) {
    if (10000 < x*x+y*y) return 0;              /* infinity */
    k = j_bbox(&a, &b, &c, &d, x, y, eps);      /* compute j(z), j'(z) */
    a = a - cx;
    b = b - cy;
    delta = c*c + d*d;
    dx = (a*c + b*d)/delta;
    dy = (b*c - a*d)/delta;
    if ((dx*dx+dy*dy) < eps*eps) return j;
    x = x - dx;
    y = y - dy;
    }
return  0;
}

int     j_bbox(jx, jy, djx, djy, x, y, eps)
double *jx, *jy, *djx, *djy;
double  x, y, eps;
{
double  tx, ty, ux, uy, delta;
int     i;
*jx = tx = 1.0; *jy = *djx = *djy = ty = 0.0;
for (i=1; i<bailout; i++) {
    delta = 4*i*(i+xa[0]);
    ux = (tx*x-ty*y)/delta;
    uy = (tx*y+ty*x)/delta;
    *jx += (tx = (ux*x-uy*y));
    *jy += (ty = (ux*y+uy*x));
    *djx += 2*i*ux;
    *djy += 2*i*uy;
    if (ux*ux+uy*uy < eps*eps) break;
    }
return i;
}

/* dynamic system z<-F(z) with global parameters eps, bailout
*   F(z) = 2*z^3+b/(3*x^2-a)
*   iterate z <- z-f(z)/f'(z) until |f(z)/f'(z)|<eps
*   here we set z = x+iy
*   start the iteration with cy, cx
*/

int     nr_cubic(cy, cx)
double  cy, cx;
{
double  x = cx, y = cy;
double  x1, x2, y1, y2, nx, ny, dx, dy, delta;
int     j;
for(j = 0; j < bailout; j++) {
    if (1e16 < x*x+y*y) return 0;      /* infinity */
    x2 = x*x; y2 = y*y;
    dx = 3*(x2-y2) - xa[0];
    dy = 6*x*y;
    delta =dx*dx+dy*dy;
    if (delta < eps * eps) return 0;
    nx = xa[1] + 2*x*(x2-3*y2);
    ny = 2*y*(3*x2-y2);
    x1 = (nx*dx+ny*dy)/delta;
    y1 = (ny*dx-nx*dy)/delta;
    dx = x-x1;
    dy = y-y1;
    delta = dx*dx + dy*dy;
    if (delta < eps*eps) return j;
    x = x1;
    y = y1;
    }
return 0;
}

/* dynamic system z<-F(z) with global parameters eps, bailout
*   F(z) = 2*z^3-b+Y^Y/(3*z^2+a)
*   iterate z <- z-f(z)/f'(z) until |f(z)/f'(z)|<eps
*   here we set z = x+iy
*   start the iteration with cy, cx
*/

int     nr_ecc(cy, cx)
double  cy, cx;
{
double  x = cx, y = cy;
double  reysq = cx*cx - cy*cy;
double  imysq = 2*cx*cy;
double  x1, x2, y1, y2, nx, ny, dx, dy, delta;
int     j;
for(j = 0; j < bailout; j++) {
    if (1e16 < x*x+y*y) return 0;      /* infinity */
    x2 = x*x; y2 = y*y;
    dx = 3*(x2-y2) + xa[0];
    dy = 6*x*y;
    delta =dx*dx+dy*dy;
    if (delta < eps*eps) return 0;
    nx = xa[1] + 2*x*(x2-3*y2) - reysq;
    ny = 2*y*(3*x2-y2) - imysq;
    x1 = (nx*dx+ny*dy)/delta;
    y1 = (ny*dx-nx*dy)/delta;
    dx = x-x1;
    dy = y-y1;
    delta = dx*dx + dy*dy;
    if (delta < eps*eps) return j;
    x = x1;
    y = y1;
    }
return 0;
}




